#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <esd.h>
#include "port.h"
#include "apu.h"
#include "soundux.h"

int g_VM=1;

 // Snes9x's Mixer
extern void S9xMixSamples (unsigned char *, int);

 // Mixes Per Second
#define MPS 100
 // Buffers Per Second (Keep this low for windows)
#define BPS 4
 // Mixes Per Buffer
#define MPB (MPS/BPS)
 // Clocks Per Mix (the SPC is 2.048 Mhz)
#define CPM (2048000/MPS)
 // How many 32 Clocks Per Mix
#define C32PM (2048000/MPS)/32

 // Bytes Per Mix
int bpm=0;
 // Bytes Per Buffer
int bpb=0;
 // Samples Per Mix (double for stereo)
int spm=0;

int hwo=-1;
unsigned char *buf;
BOOL playing=FALSE;

SAPURegisters BackupAPURegisters;
unsigned char BackupAPURAM[65536];
unsigned char BackupAPUExtraRAM[64];
unsigned char BackupDSPRAM[128];

SDSPStream dspstream;
unsigned long curdspbuffer=0;
char DSP=0;

// The callback

void DoTimer(){
	if (APU.TimerEnabled [2])
	{
		APU.Timer [2]++;
		if (APU.Timer [2] >= APU.TimerTarget [2])
		{
			IAPU.RAM [0xff] = (IAPU.RAM [0xff] + 1) & 0xf;
			APU.Timer [2] = 0;
#ifdef SPC700_SHUTDOWN          
			IAPU.WaitCounter++;
			IAPU.APUExecuting = TRUE;
#endif          

		}
	}
	if (IAPU.TimerErrorCounter>=8)
	{
		IAPU.TimerErrorCounter=0;
		if (APU.TimerEnabled [0])
		{
			APU.Timer [0]++;
			if (APU.Timer [0] >= APU.TimerTarget [0])
			{
				IAPU.RAM [0xfd] = (IAPU.RAM [0xfd] + 1) & 0xf;
				APU.Timer [0] = 0;
#ifdef SPC700_SHUTDOWN          
				IAPU.WaitCounter++;
				IAPU.APUExecuting = TRUE;
#endif          

			}
		}
		if (APU.TimerEnabled [1])
		{
			APU.Timer [1]++;
			if (APU.Timer [1] >= APU.TimerTarget [1])
			{
				IAPU.RAM [0xfe] = (IAPU.RAM [0xfe] + 1) & 0xf;
				APU.Timer [1] = 0;
#ifdef SPC700_SHUTDOWN          
				IAPU.WaitCounter++;
				IAPU.APUExecuting = TRUE;
#endif          
			}
		}
	}
}

// Play SPC

void PlaySPC(){
	int i;
	unsigned char *p, *buftop = buf+(MPB*bpm);

	playing = TRUE;
	
	while(playing){
		for(p=buf; p<buftop; p+=bpm){
			// APU_LOOP
			for(int c=C32PM; c; c--){
				for(int ic=32; ic; ic--){
					APU_EXECUTE1();
				}
				IAPU.TimerErrorCounter++;
				DoTimer();
			}
			S9xMixSamples((unsigned char *)(p),spm);
		}
		write(hwo, buf, bpb);
	}
}

// Pause SPC

void PauseSPC()
{
	if(hwo&&playing){
		playing=FALSE;
	}
}

// Open Sound Device

BOOL OpenSoundDevice (int rate, int bits, int channels)
{
	int b, fmt = (channels == 2 ? ESD_STEREO : ESD_MONO) |
	             (bits == 16 ? ESD_BITS16 : ESD_BITS8);

	if( hwo == -1 ){
		hwo = esd_play_stream_fallback( fmt, rate, NULL, NULL );
		if( hwo == -1 ) return FALSE;
	}

	//SPC mixer information
	so.stereo = (channels==2)?TRUE:FALSE;
	so.sixteen_bit = (bits==16)?TRUE:FALSE;
	so.playback_rate = rate;
	so.buffer_size = rate/MPS;
	so.encoded = FALSE;
	so.mute_sound = FALSE;

	//Buffer mixing information
	bpb=rate*channels*(bits==16?2:1)/BPS;
	bpm=rate*channels*(bits==16?2:1)/MPS;
	spm=rate*channels/MPS;

	buf=(unsigned char *)malloc(bpb);

	return TRUE;
}

// Close Sound Device

void CloseSoundDevice(){
	int i;

	if(hwo!=-1){
		close(hwo);
		playing=FALSE;
	}
}

// Restore SPC Load State

void FixEnvelope (int channel, uint8 gain, uint8 adsr1, uint8 adsr2);
void StartSample (int channel);
void S9xSetAPUDSP (uint8 byte);

void RestoreSPC(){
	int i;

	APURegisters.PC = BackupAPURegisters.PC;
	APURegisters.YA.B.A = BackupAPURegisters.YA.B.A;
	APURegisters.X = BackupAPURegisters.X;
	APURegisters.YA.B.Y = BackupAPURegisters.YA.B.Y;
	APURegisters.P = BackupAPURegisters.P;
	APURegisters.S = BackupAPURegisters.S;
	memcpy(IAPU.RAM,BackupAPURAM,65536);
	memcpy(APU.ExtraRAM,BackupAPUExtraRAM,64);
	memcpy(APU.DSP,BackupDSPRAM,128);

	for(i=0;i<4;i++){
		APU.OutPorts [i] = IAPU.RAM[0xf4+i];
	}
	IAPU.TimerErrorCounter=0;
	for (i=0;i<3;i++){
		if (IAPU.RAM[0xfa+i] == 0)
			APU.TimerTarget[i] = 0x100;
		else
			APU.TimerTarget[i] = IAPU.RAM[0xfa+i];
	}
	S9xSetAPUControl(IAPU.RAM[0xf1]);
	S9xSetSoundMute (FALSE);
	IAPU.PC = IAPU.RAM + APURegisters.PC;
	S9xAPUUnpackStatus();
	if (APUCheckDirectPage ())
		IAPU.DirectPage = IAPU.RAM + 0x100;
	else
		IAPU.DirectPage = IAPU.RAM;
	IAPU.APUExecuting = TRUE;

	S9xFixSoundAfterSnapshotLoad();

	S9xSetFrequencyModulationEnable(APU.DSP[APU_PMON]);
	S9xSetMasterVolume(APU.DSP[APU_MVOL_LEFT],APU.DSP[APU_MVOL_RIGHT]);
	S9xSetEchoVolume(APU.DSP[APU_EVOL_LEFT],APU.DSP[APU_EVOL_RIGHT]);
	uint8 mask = 1;
	int type;
	for(i=0;i<8;i++,mask<<=1){
		FixEnvelope (i,
			APU.DSP[APU_GAIN+(i<<4)],
			APU.DSP[APU_ADSR1+(i<<4)],
			APU.DSP[APU_ADSR2+(i<<4)]);
		S9xSetSoundVolume (i,
			APU.DSP[APU_VOL_LEFT+(i<<4)],
			APU.DSP[APU_VOL_RIGHT+(i<<4)]);
        S9xSetSoundFrequency (i,
			((APU.DSP[APU_P_LOW+(i<<4)]+(APU.DSP[APU_P_HIGH+(i<<4)]<< 8))&FREQUENCY_MASK)*8);
		if (APU.DSP [APU_NON] & mask){
			type = SOUND_NOISE;
		}else{
			type = SOUND_SAMPLE;
		}
		S9xSetSoundType (i, type);
		if ((APU.DSP[APU_KON] & mask) != 0){
			APU.KeyedChannels |= mask;
			StartSample (i);
		}
	}
	/*unsigned char temp=IAPU.RAM[0xf2];
	for(i=0;i<128;i++){
		IAPU.RAM[0xf2]=i;
		S9xSetAPUDSP(APU.DSP[i]);
	}
	IAPU.RAM[0xf2]=temp;*/
	curdspbuffer=0;
}

// Load SPC file

BOOL LoadSPC(char *fname)
{
	int fd;

	fd = open(fname, O_RDONLY);
	if( fd == -1 ) return FALSE;
	
	S9xResetAPU();

	lseek(fd, 0x25, SEEK_SET);
	read(fd, &BackupAPURegisters.PC,2);
	read(fd, &BackupAPURegisters.YA.B.A,1);
	read(fd, &BackupAPURegisters.X,1);
	read(fd, &BackupAPURegisters.YA.B.Y,1);
	read(fd, &BackupAPURegisters.P,1);
	read(fd, &BackupAPURegisters.S,1);
	lseek(fd, 0x100, SEEK_SET);
	read(fd, BackupAPURAM,65536);
	read(fd, BackupDSPRAM,128);
	lseek(fd, 64, SEEK_CUR);
	read(fd, BackupAPUExtraRAM,64);

	close(fd);
	RestoreSPC();
		
	DSP=0;
	return TRUE;
}
